//---------------------------------------------------------------------------

#pragma hdrstop

#include "markov.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

//#define DEBUG

int N_PREFIX_WORD = 2; //  

/*    */
char *GetWord( FILE *f ){
 const N_LEX = 10000;
 static char buf[N_LEX+1];
 int k = 0 ,ch;

 while ( (ch = fgetc(f)) != EOF  &&  isspace(ch) ); //skip spaces
 if(ch == EOF) return NULL;

 buf[k++] = ch;
 while ( (ch = fgetc(f)) != EOF  &&  !isspace(ch) )  // read word
   if( k < N_LEX ) buf[k++] = ch;

 buf[k] = '\0';

 if(k == 0) return NULL;
 return buf;

}

/*        ,   - NULL */
TWordPtrArray FileToWords( char fname[],
                     char* (*GetWord)( FILE *f ) //  
                  )
{
  FILE *f;
  int cnt_words = 0;
  TWord **array = NULL;



  if( (f = fopen(fname,"r")) != NULL ){

     while( GetWord(f) ) cnt_words++;   //- 
     fclose(f);

     array = (TWord**)calloc( cnt_words + 1, sizeof(TWord*));
     if(array)
     {
       char *word;
       int k = 0;

       f = fopen(fname,"r");
       while( (word = GetWord(f)) != NULL )
         array[k++] =  TWordNew( strdup(word) );
       fclose(f);
     }
  }
  return array;
}

TWord *TWordNew( char s[] ){
   TWord *p = (TWord*)malloc( sizeof(TWord));
   if( p )
   {
     unsigned key = 0;
     unsigned char *str = (unsigned char*)s;
     while(*str){
        key = key*31 + *str;
        str++;
     }
     p->key = key;
     p->len = str - s;
     p->word = s;
   }
   return p;
}


// -   N_PREFIX_WORD 
unsigned TPrefixHashKey( TPrefix *node ){
  unsigned key = 0;
  TWord **word = node->pref_array;
  int n = 0;
  while(n < N_PREFIX_WORD && *word){
     key ^= (*word)->key;
     word++;
     n++;
  }
  return key;
}

TPrefix* THashPrefixLook( THashPrefix* h, TPrefix* node ){
  unsigned node_key = TPrefixHashKey(node);
  TPrefix *entry = h->hash[node_key & (HASH_SIZE-1)];
  while(entry){
   if(TPrefixHashKey(entry) == node_key){ //    
     TWord **w1 = node->pref_array;
     TWord **w2 = entry->pref_array;
     int n = 0;
     while(n < N_PREFIX_WORD  &&         //  
          *w1 && *w2 &&
           strcmp((*w1)->word, (*w2)->word) == 0){

       n++;
       w1++;
       w2++;
     }
     if(n == N_PREFIX_WORD) return entry;  // 
   }
   entry = entry->next;
  }
  return NULL;                            //  
}

TPrefix* THashPrefixInsert( THashPrefix* h, TPrefix* node ){
  TPrefix **entry = &h->hash[TPrefixHashKey(node) & (HASH_SIZE-1)];

  node->next = *entry;
  *entry = node;

  node->nextMainList = h->list;
  h->list = node;
  h->cnt++;

  return *entry;
}

void THashPrefixPrint( THashPrefix* h ){
  TPrefix *pr;
  TSuffix *suf;

  for(pr = h->list; pr; pr = pr->nextMainList)
  {
    TWord **w;
    int k = 0;
    char *s;
    for(k = 0, w =  pr->pref_array;  k < N_PREFIX_WORD && *w; k++,w++)
    {
       s = (*w)->word;
       printf("%s ", s);
    }
    printf("- ");
    for(suf = pr->suffix_list; suf; suf = suf->next)
    {
       s = (*suf->suf)->word;
       printf("%s(%d) ", s, suf->cnt );
    }
    printf("\n");
  }
}

TPrefix* THashPrefixAt( THashPrefix* h, int index ){
  TPrefix *pr;
  int cnt = 0;
  for(pr = h->list; pr; pr = pr->nextMainList)
   if(++cnt == index) return pr;

  return 0;
}


TPrefix* TPrefixNew(TWord **pref_array,TSuffix *suffix_list,TPrefix *next)
{
   TPrefix *p = (TPrefix*)malloc(sizeof(TPrefix));
   if(p){
     p->pref_array = pref_array;
     p->suffix_list = suffix_list;
     p->next = next;
     p->nextMainList = NULL;
   }
   return p;
}

TSuffix* TSuffixNew(TWord **suf,TSuffix* next,int cnt){

  TSuffix * p = (TSuffix*)calloc(1,sizeof(TSuffix));
  if(p){
    p->suf = suf;
    p->next = next;
    p->cnt = cnt;
  }
  return p;
}


int GenMarkovText( char fname_source[],int max_words_output )
{
  //   !!!
  TWordPtrArray words = FileToWords(fname_source, GetWord);
  TWordPtrArray output_words = (TWord**)calloc(max_words_output+1, sizeof(TWord));
  THashPrefix *hash = THashPrefixNew();
  int cnt_output_words = 0;


  if( words && output_words && hash ){
     int i;
     TPrefix *tmp_pref = TPrefixNew(0,0,0);


     //     N_PREFIX_WORD   +   
     // (     )
     for( i = 0; words[i]; i++ )
       if(i >= N_PREFIX_WORD){
          TPrefix *pref;
          tmp_pref->pref_array = words + (i - N_PREFIX_WORD);
          pref = THashPrefixLook(hash, tmp_pref);
          if(pref==NULL){
             pref = THashPrefixInsert(hash,
                              TPrefixNew( tmp_pref->pref_array,0,0)
                              );
          }
          if( pref ){
            TSuffix *suf;

            for(suf = pref->suffix_list; suf; suf = suf->next)
              if( strcmp( (*suf->suf)->word , words[i]->word) == 0 )
              {
                suf->cnt++;
                break;
              }

            if(suf==NULL)
                pref->suffix_list = TSuffixNew(words + i,pref->suffix_list,1);

          }//if pref
       }
#ifdef DEBUG
    THashPrefixPrint(hash);
    return 0;
#endif

       //    
        srand(time(0));
        do{
          TPrefix *p = THashPrefixAt(hash, rand()%hash->cnt);
          TWord **w;
          if(p==NULL) goto done;
          for(w = p->pref_array;
              *w && cnt_output_words < N_PREFIX_WORD && cnt_output_words < max_words_output;
              w++, cnt_output_words++)
              output_words[cnt_output_words] = *w;
          if( cnt_output_words < N_PREFIX_WORD ) goto done;
        }while(0);


        //  
        while(cnt_output_words < max_words_output){
          TPrefix* find_pref;
          TSuffix *suf;
          tmp_pref->pref_array = output_words + cnt_output_words - N_PREFIX_WORD;
          find_pref = THashPrefixLook(hash, tmp_pref);
          if( find_pref == NULL ) goto done;
          suf = find_pref->suffix_list;
          if(suf==NULL) goto done;
          do{
            int is_find_suffix = 0;
            int summ_cnt = 0, probability;
            TSuffix *p;
            for(p = suf; p; p = p->next)
              summ_cnt += p->cnt;
            probability = rand()%summ_cnt;
            summ_cnt = 0;
            for(p = suf; p; p = p->next){

              if( probability >= summ_cnt  &&
                  probability <  summ_cnt + p->cnt )
              {
                output_words[cnt_output_words++] = *p->suf;
                is_find_suffix = 1;
                break;
              }
              summ_cnt += p->cnt;
            }
            if(is_find_suffix == 0) goto done;
          }while(0);
        }//        while(cnt_output_words < max_words_output)

  } //endif words

done:

  do{       //     
    int j;
    for( j = 0; j < cnt_output_words; j++)
     printf("%s ", output_words[j]->word);

  }while(0);


  return cnt_output_words;
}





